# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import absolute_import, division, print_function

from inspire_json_merger.comparators import AuthorComparator


def get_number_of_author_matches(x_authors, y_authors):
    """Return the number of matches between two lists of authors.

    Args:
        x_authors (list(dict)): a schema-compliant list of authors.
        y_authors (list(dict)): another schema-compliant list of authors.

    Returns:
        int: the number of matching authors between the two lists.

    """
    return len(AuthorComparator(x_authors, y_authors).matches)


def compute_author_match_score(x_authors, y_authors):
    """Return the matching score of 2 given lists of authors.

    Args:
        x_authors (list(dict)): first schema-compliant list of authors.
        y_authors (list(dict)): second schema-compliant list of authors.

    Returns:
        float: matching score of authors; deafult is 0.5 .

    """
    if len(x_authors) <= 1 or len(y_authors) <= 1:
        return 0.5

    matches = get_number_of_author_matches(x_authors, y_authors)
    max_length = max(len(x_authors), len(y_authors))

    return matches / float(max_length)


def compute_jaccard_index(x_set, y_set):
    """Return the Jaccard similarity coefficient of 2 given sets.

    Args:
        x_set (set): first set.
        y_set (set): second set.

    Returns:
        float: Jaccard similarity coefficient.

    """
    if not x_set or not y_set:
        return 0.0

    intersection_cardinal = len(x_set & y_set)
    union_cardinal = len(x_set | y_set)

    return intersection_cardinal / float(union_cardinal)


def get_tokenized_title(title):
    """Return the tokenised title.

    The title is lowercased and split on the spaces. Then, duplicate tokens are removed by adding the tokens to a set.

    Args:
        title (string): a title.

    Returns:
        set: contains the resulting tokens.

    """
    return set(title.lower().split())



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Matcher API."""

from __future__ import absolute_import, division, print_function

from flask import current_app
from werkzeug.utils import import_string

from invenio_search import current_search_client as es

from .core import compile


def _get_validator(validator_param):
    if callable(validator_param):
        return validator_param

    try:
        validator = import_string(validator_param)
    except (KeyError, ImportError):
        current_app.logger.debug('No validator provided. Falling back to the default validator.')
        validator = import_string('inspire_matcher.validators:default_validator')

    return validator


def match(record, config=None):
    """Given a record, yield the records in INSPIRE most similar to it.

    This method can be used to detect if a record that we are ingesting as a
    submission or as an harvest is already present in the system, or to find
    out which record a reference should be pointing to.
    """
    if config is None:
        current_app.logger.debug('No configuration provided. Falling back to the default configuration.')
        config = current_app.config['MATCHER_DEFAULT_CONFIGURATION']

    try:
        algorithm, doc_type, index = config['algorithm'], config['doc_type'], config['index']
    except KeyError as e:
        raise KeyError('Malformed configuration: %s.' % repr(e))

    source = config.get('source', [])

    for i, step in enumerate(algorithm):
        try:
            queries = step['queries']
        except KeyError:
            raise KeyError('Malformed algorithm: step %d has no queries.' % i)

        validator = _get_validator(step.get('validator'))

        for j, query in enumerate(queries):
            try:
                body = compile(query, record)
            except Exception as e:
                raise ValueError('Malformed query. Query %d of step %d does not compile: %s.' % (j, i, repr(e)))

            if not body:
                continue

            if source:
                result = es.search(index=index, doc_type=doc_type, body=body, _source=source)
            else:
                result = es.search(index=index, doc_type=doc_type, body=body)

            for hit in result['hits']['hits']:
                if validator(record, hit):
                    yield hit



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import absolute_import, division, print_function

import os
import warnings

from inspire_utils.helpers import force_list
from inspire_utils.record import get_value


def compile(query, record):
    type_ = query['type']

    if type_ == 'exact':
        return _compile_exact(query, record)
    elif type_ == 'fuzzy':
        return _compile_fuzzy(query, record)
    elif type_ == 'nested':
        return _compile_nested(query, record)

    raise NotImplementedError(type_)


def _compile_exact(query, record):
    if 'match' in query:
        query['path'] = query.get('path', query['match'])
        warnings.warn('The "match" key is deprecated. Use "path" instead.', DeprecationWarning)

    if 'search' in query:
        query['search_path'] = query.get('search_path', query['search'])
        warnings.warn('The "search" key is deprecated. Use "search_path" instead.', DeprecationWarning)

    path, search_path = query['path'], query['search_path']

    collections = query.get('collections', [])

    values = force_list(get_value(record, path))
    if not values:
        return

    result = {
        'query': {
            'bool': {
                'should': [],
            },
        },
    }

    if collections:
        result['query']['bool']['minimum_should_match'] = 1
        result['query']['bool']['filter'] = {'bool': {'should': []}}

        for collection in collections:
            result['query']['bool']['filter']['bool']['should'].append({
                'match': {
                    '_collections': collection,
                },
            })

    for value in values:
        result['query']['bool']['should'].append({
            'match': {
                search_path: value,
            },
        })

    return result


def _compile_fuzzy(query, record):
    clauses = query['clauses']

    result = {
        'min_score': 1,
        'query': {
            'dis_max': {
                'queries': [],
                'tie_breaker': 0.3,
            },
        },
    }

    for clause in clauses:
        path = clause['path']

        boost = clause.get('boost', 1)

        values = get_value(record, path)
        if not values:
            continue

        path = path.split('[')[0]
        if '.' in path:
            raise ValueError('the "path" key can\'t contain dots')
        # TODO: This query should be refined instead of relying on validation to filter out irrelevant results.
        result['query']['dis_max']['queries'].append({
            'more_like_this': {
                'boost': boost,
                'docs': [
                    {
                        'doc': {
                            path: values,
                        },
                    },
                ],
                'max_query_terms': 25,
                'min_doc_freq': 1,
                'min_term_freq': 1,
            },
        })

    if not result['query']['dis_max']['queries']:
        return

    return result


def _compile_nested(query, record):
    paths, search_paths = query['paths'], query['search_paths']

    if len(paths) != len(search_paths):
        raise ValueError('paths and search_paths must be of the same length')

    common_path = _get_common_path(search_paths)
    if not common_path:
        raise ValueError('search_paths must share a common path')

    result = {
        'query': {
            'nested': {
                'path': common_path,
                'query': {
                    'bool': {
                        'must': [],
                    },
                },
            },
        },
    }

    for path, search_path in zip(paths, search_paths):
        value = get_value(record, path)
        if not value:
            return

        result['query']['nested']['query']['bool']['must'].append({
            'match': {
                search_path: value,
            },
        })

    return result


def _get_common_path(paths):
    return '.'.join(os.path.commonprefix([path.split('.') for path in paths]))



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2017 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Matcher validators."""

from __future__ import absolute_import, division, print_function

from itertools import product

from inspire_utils.record import get_value

from .utils import (
    compute_author_match_score,
    compute_jaccard_index,
    get_tokenized_title,
)


def default_validator(record, result):
    return True


def authors_titles_validator(record, result):
    """Compute a validation score for the possible match.

    The score is based on a similarity score of the authors sets and the maximum Jaccard index found between 2 titles:
    one from the record and one from the result title sets.

    If the computed score is higher than 0.5, then the match is valid, otherwise it is not.

    Args:
        record (dict): the given record we are trying to match with similar ones in INSPIRE.
        result (dict): possible match returned by the ES query that needs to be validated.

    Returns:
        bool: validation decision.

    """
    record_authors = get_value(record, 'authors', [])
    result_authors = get_value(result, '_source.authors', [])

    author_score = compute_author_match_score(record_authors[:5], result_authors[:5])

    title_max_score = 0.5
    record_titles = get_value(record, 'titles.title', [])
    result_titles = get_value(result, '_source.titles.title', [])

    for cartesian_pair in product(record_titles, result_titles):
        record_title_tokens = get_tokenized_title(cartesian_pair[0])
        result_title_tokens = get_tokenized_title(cartesian_pair[1])
        current_title_jaccard = compute_jaccard_index(record_title_tokens, result_title_tokens)

        if current_title_jaccard > title_max_score:
            title_max_score = current_title_jaccard

    return (author_score + title_max_score) / 2 > 0.5
